import { VERSIONS } from "./versions";
import { TransformResponse } from "../../../utils/transform-response";
import { GetClassMethods } from "../utils/reflection";
import { Router } from "express";
import { ChunithmVersions } from "lib/constants/game-versions";
import CreateLogCtx from "lib/logger/logger";
import { RequestLoggerMiddleware } from "lib/middleware/request-logger";
import { Config, Environment } from "lib/setup/config";
import getRawBody from "raw-body";
import { IsHexadecimalString } from "utils/misc";
import { ZodError } from "zod";
import { fromZodError } from "zod-validation-error";
import { createCipheriv, createDecipheriv, pbkdf2Sync } from "crypto";
import { promisify } from "util";
import zlib, { createDeflate, inflate } from "zlib";
import type { ErrorRequestHandler, RequestHandler } from "express";
import type { Transform } from "stream";

const logger = CreateLogCtx(__filename);

const inflateAsync = promisify(inflate);

const router: Router = Router({ mergeParams: false });

router.use(async (req, res, next) => {
	if ((req.headers["content-length"] ?? 0) === 0) {
		next();
		return;
	}

	// /SDHD/210/UpsertUserAllApi
	const splitPath = req.path.split("/");
	const version = Number(splitPath[2]);
	const endpoint = splitPath.at(-1);

	if (!endpoint || Number.isNaN(version)) {
		next();
		return;
	}

	const isEncryptedRequest = endpoint.length === 32 && IsHexadecimalString(endpoint);

	if (
		!isEncryptedRequest &&
		Config.CHUNITHM_CONFIG.CRYPTO.ENCRYPTED_ONLY &&
		version >= ChunithmVersions.CRYSTAL_PLUS
	) {
		logger.error(
			"Rejecting unencrypted request, because the server is configured to accept only encrypted requests.",
			{ version, endpoint }
		);

		return res.status(200).send({ stat: "0" });
	}

	const transformations: Array<Transform> = [
		createDeflate({ flush: zlib.constants.Z_SYNC_FLUSH }),
	];
	let decrypted: Buffer | null = null;

	if (isEncryptedRequest) {
		const keys = Config.CHUNITHM_CONFIG.CRYPTO.KEYS?.[version];

		if (!keys) {
			logger.error("No keys were configured for the requested version.", {
				version,
				endpoint,
			});
			return res.status(200).send({ stat: "0" });
		}

		// Mostly as a convenience, so versions PARADISE LOST and older don't have to provide
		// the path salt.
		const userAgent = req.headers["user-agent"];

		if (version <= ChunithmVersions.PARADISE && userAgent) {
			const unhashedEndpoint = userAgent.split("#")[0];

			if (unhashedEndpoint) {
				logger.debug("Resolved to unhashed endpoint from User-Agent.", {
					unhashedEndpoint,
				});
				req.url = req.url.replace(endpoint, unhashedEndpoint);
			} else {
				logger.warn(
					"User-Agent did not contain the unhashed endpoint for version PARADISE (LOST) or older?",
					{ version, userAgent }
				);
			}
		}

		const keyString = keys[0];
		const ivString = keys[1];

		if (!keyString || !ivString) {
			logger.error("Key or IV was not provided for the requested version.", {
				version,
				key: keyString,
				iv: ivString,
			});

			return res.status(200).send({ stat: "0" });
		}

		const key = Buffer.from(keyString, "hex");
		const iv = Buffer.from(ivString, "hex");
		const cipher = createDecipheriv("aes-256-cbc", key, iv);
		const body = await getRawBody(req, { limit: "20mb" });

		res.locals.key = key;
		res.locals.iv = iv;
		decrypted = Buffer.concat([cipher.update(body), cipher.final()]);

		transformations.push(createCipheriv("aes-256-cbc", key, iv));
	} else {
		decrypted = await getRawBody(req, { limit: "20mb" });
	}

	const decompressed = await inflateAsync(decrypted);

	// eslint-disable-next-line require-atomic-updates
	req.body = JSON.parse(decompressed.toString("utf-8"));

	TransformResponse(res, ...transformations);
	res.locals.transformationsSet = true;

	next();
});

router.use((req, _, next) => {
	if (Environment.nodeEnv === "dev" && req.header("x-debug") !== undefined) {
		next();
		return;
	}

	// Get express.json() to recognize
	req.headers["content-encoding"] = "deflate";

	next();
});

router.use((req, res, next) => {
	// Always mount an empty req body. We operate under the assumption that req.body is
	// always defined as atleast an object.
	if (req.method !== "GET" && (typeof req.body !== "object" || req.body === null)) {
		req.body = {};
	}

	// req.safeBody *is* just a type-safe req.body!
	req.safeBody = req.body as Record<string, unknown>;

	next();
});

router.use(RequestLoggerMiddleware);

for (const Title of VERSIONS) {
	const inst = new Title();

	const endpointSaltString = Config.CHUNITHM_CONFIG.CRYPTO.KEYS?.[inst.version]?.[2];
	const iterations = inst.numericVersion >= ChunithmVersions.SUN ? 70 : 44;

	const key = Config.CHUNITHM_CONFIG.CRYPTO.KEYS?.[inst.version]?.[0];
	const iv = Config.CHUNITHM_CONFIG.CRYPTO.KEYS?.[inst.version]?.[1];

	if (
		Config.CHUNITHM_CONFIG.CRYPTO.ENCRYPTED_ONLY &&
		inst.numericVersion >= ChunithmVersions.CRYSTAL_PLUS &&
		(!key || !iv)
	) {
		logger.warn(
			`Disabling ${inst.gameCode} v${inst.version}, because ENCRYPTED_ONLY was enabled but no keys are provided.`,
			{
				gameCode: inst.gameCode,
				version: inst.version,
			}
		);
		continue;
	}

	if (
		Config.CHUNITHM_CONFIG.CRYPTO.ENCRYPTED_ONLY &&
		inst.numericVersion >= ChunithmVersions.NEW &&
		!endpointSaltString
	) {
		logger.warn(
			`Disabling ${inst.gameCode} v${inst.version}, because ENCRYPTED_ONLY was enabled but no endpoint salt was provided.`,
			{
				gameCode: inst.gameCode,
				version: inst.version,
			}
		);
	}

	const subrouter: Router = Router({ mergeParams: false });
	const methods = GetClassMethods(Title).filter((m) => m.startsWith("handle_"));
	const endpointSalt = endpointSaltString ? Buffer.from(endpointSaltString, "hex") : null;

	for (const method of methods) {
		const methodName = method.split("_")[1];

		if (!methodName) {
			logger.warn(
				"Method starts with 'handle_', but cannot get the part after the underscore?",
				{
					method,
				}
			);
			continue;
		}

		const impl = (inst as unknown as Record<string, unknown>)[method] as RequestHandler;
		const implWrapper: RequestHandler = async (req, res, next) => {
			// eslint-disable-next-line @typescript-eslint/no-confusing-void-expression
			await Promise.resolve(impl.call(inst, req, res, next));
		};

		subrouter.post(`/${methodName}`, implWrapper);

		if (endpointSalt) {
			const hash = pbkdf2Sync(methodName, endpointSalt, iterations, 16, "sha1").toString(
				"hex"
			);

			subrouter.post(`/${hash}`, implWrapper);
		}
	}

	router.use(`/${inst.gameCode}/${inst.version}/${inst.servletName}`, subrouter);

	// TODO: Add to a title registry
}

const ERR_HANDLER: ErrorRequestHandler = (err, req, res, _next) => {
	// eslint-disable-next-line cadence/no-instanceof
	const reportedErr = err instanceof ZodError ? fromZodError(err).message : err;

	logger.error("Error propagated to title router.", {
		err: reportedErr,
		url: req.originalUrl,
	});

	if (!res.locals.transformationsSet) {
		const transformations: Array<Transform> = [
			createDeflate({ flush: zlib.constants.Z_SYNC_FLUSH }),
		];

		if (res.locals.key && res.locals.iv) {
			transformations.push(createCipheriv("aes-256-cbc", res.locals.key, res.locals.iv));
		}

		TransformResponse(res, ...transformations);
	}

	return res.status(200).send({ returnCode: "0" });
};

router.use(ERR_HANDLER);

export default router;
