import { ChunkLength, ToBuffer } from "../servers/titles/utils/buffer";
import onHeaders from "on-headers";
import type { Response } from "express";

/**
 * Pipe the response through a series of transformations.
 *
 * Inspired by [node.js compression middleware](https://www.npmjs.com/package/compression).
 * @param res The response to transform.
 * @param transformers stream.Transform items.
 */
export function TransformResponse(res: Response, ...transformers: Array<NodeJS.ReadWriteStream>) {
	const head = transformers[0];
	const tail = transformers.at(-1);

	if (!head || !tail) {
		throw new Error("At least one stream.Transform must be provided");
	}

	if (transformers.length > 1) {
		for (let i = 0; i < transformers.length - 1; i++) {
			const previous = transformers[i];
			const next = transformers[i + 1];

			if (!previous || !next) {
				throw new Error("The provided transformers array have an undefined item.");
			}

			previous.pipe(next);
		}
	}

	let ended = false;

	/* eslint-disable @typescript-eslint/unbound-method */
	const _on = res.on;
	const _write = res.write;
	const _end = res.end;
	/* eslint-enable @typescript-eslint/unbound-method */

	/* eslint-disable prefer-arrow-callback */
	res.write = function write(chunk, encoding) {
		if (ended) {
			return false;
		}

		// @ts-expect-error Typical JavaScript monkey patching.
		if (!this._header && !this.headersSent) {
			this.writeHead(this.statusCode);
		}

		const enc = typeof encoding === "function" ? undefined : encoding;

		return head.write(ToBuffer(chunk, enc));
	};

	// @ts-expect-error More monkey patching shenanigans. It works, we'll live.
	res.end = function end(chunk, encoding) {
		if (ended) {
			return false;
		}

		const enc = typeof encoding === "function" ? undefined : encoding;

		// @ts-expect-error how do express people live like this?
		if (!this._header && !this.headersSent) {
			// estimate the length
			if (!this.getHeader("Content-Length")) {
				this.setHeader("Content-Length", ChunkLength(chunk, enc));
			}

			this.writeHead(this.statusCode);
		}

		// mark ended
		ended = true;

		// write Buffer for Node.js 0.8
		if (chunk) {
			// @ts-expect-error hhh
			head.end(ToBuffer(chunk, encoding));
		} else {
			head.end();
		}

		return this;
	};

	// @ts-expect-error zzz
	res.on = function on(event, listener) {
		if (event !== "drain") {
			return _on.call(this, event, listener);
		}

		return head.on(event, listener);
	};
	/* eslint-enable prefer-arrow-callback */

	onHeaders(res, () => {
		res.removeHeader("Content-Length");

		tail.on("data", (chunk) => {
			// @ts-expect-error ugggh i hate monkeys
			if (!_write.call(res, chunk)) {
				head.pause();
			}
		});

		tail.on("end", () => {
			// @ts-expect-error ugggh i hate monkeys
			_end.call(res);
		});

		_on.call(res, "drain", () => {
			head.resume();
		});
	});
}
